/*
** Copyright (c) 2000-2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: binary.c,v 1.9 2002/06/04 06:37:23 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/file.h>

#include "ember.h"
#include "config/config.h"

#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif

/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

static void intToBuf(cp,val)
        u_char  *cp;
        int     val;
{
        *cp++ = (unsigned int)(val & 0x000000ff);
        *cp++ = (unsigned int)(val & 0x0000ff00) >> 8;
        *cp++ = (unsigned int)(val & 0x00ff0000) >> 16;
        *cp++ = (unsigned int)(val & 0xff000000) >> 24;
}


static int bufToInt(cp)
        u_char  *cp;
{
        int val;

        val = 0;
        val = *cp++;
        val += ((int) *cp++) << 8 ;
        val += ((int) *cp++) << 16;
        val += ((int) *cp++) << 24;
        return(val);
}

static int writeInt(fd, val)
	int	fd,
		val;
{
	char	buf[4];

	intToBuf(buf, val);
	return(write(fd, buf, 4));
}


static int readInt(fd, val)
	int	fd,
		*val;
{
	char	buf[4];

	if(read(fd, buf, 4) <= 0)
		return(-1);
	*val = bufToInt(buf);
	return(0);
}


static int eBinWriteLiterals(fd, script, base)
	int	fd;
	ember	*script;
	e_stab	*base[];
{
	int     count,
		valLength,
		length;
	e_bin	bin;
	e_stab  *curSym;
	char	*value;

	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curSym = base[count];
		while(curSym)
		{
			if (curSym->value.initialised)
			{
				value=(*curSym->value.typePtr->unpack) 
					(&curSym->value);
				valLength = strlen(value) + 1;
			}
			else
			{
				valLength = 0;	
			}
		
			bin.type = EMBER_BIN_LITERAL;	
			length = 4 + 
				4 + 
				valLength;
			intToBuf(bin.length,length);
			write(fd, &bin, sizeof(bin));

			writeInt(fd, curSym->id);
			writeInt(fd,curSym->value.type);
			if (valLength)
				write(fd, value, valLength);
			curSym = curSym->next;
                }
		count++;
	}
	return(0);
}


static int eBinWriteGlobalVariables(fd, script)
	int	fd;
	ember	*script;
{
	e_bin	bin;
	e_stab	*curSym;
	e_map	*curMap;
	int	length,
		count;;

	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curMap = script->main.variables[count];
		while(curMap)
		{
			curSym = eSymGetVariableById(script, curMap->id);
			bin.type = EMBER_BIN_GLOBAL;
			length = 4 + 
				strlen(curSym->name) + 1 +
				4;
			intToBuf(bin.length,length);
			write(fd, &bin, sizeof(bin));

			writeInt(fd, curSym->id);
			writeInt(fd,curSym->value.type);
			write(fd, curSym->name, strlen(curSym->name)+ 1);
			curMap = curMap->next;
		}
		count++;
	}
	return(0);
}

static int eBinWriteFunctionHeader(fd,function)
	int	fd;
	e_funct	*function;
{
	e_bin	bin;
	int	length;

	bin.type = EMBER_BIN_FUNCT;	
	length = 4 + strlen(function->name) + 1;
	intToBuf(bin.length,length);
	write(fd, &bin, sizeof(bin));
	
	writeInt(fd, function->id);
	write(fd, function->name, strlen(function->name) + 1);
	return(0);
}

static int eBinWriteFunctionArgs(fd,function)
	int	fd;
	e_funct	*function;
{
	e_bin	bin;
	e_arg	*curArg;
	int	length;

	curArg = function->args;
	while(curArg)
	{
		bin.type = EMBER_BIN_ARG;
		length = strlen(curArg->name) + 1 +
			4 +
			4;
		intToBuf(bin.length,length);
		write(fd, &bin, sizeof(bin));
		writeInt(fd, curArg->id);
		writeInt(fd, curArg->type);
		write(fd, curArg->name, strlen(curArg->name) + 1);

		curArg = curArg->next;
	}
	return(0);
}

static int eBinWriteFunctionVariables(fd, script, base)
	int	fd;
	ember	*script;
	e_map	*base[];
{
	e_bin	bin;
	e_stab	*curSym;
	e_map	*curMap;
	int	length,
		count;;

	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curMap = base[count];
		while(curMap)
		{
			curSym = eSymGetVariableById(script, curMap->id);
			bin.type = EMBER_BIN_VARIABLE;
			length = 4 + 
				strlen(curSym->name) + 1 +
				4;
			intToBuf(bin.length,length);
			write(fd, &bin, sizeof(bin));

			writeInt(fd, curSym->id);
			writeInt(fd,curSym->value.type);
			write(fd, curSym->name, strlen(curSym->name)+ 1);
			curMap = curMap->next;
		}
		count++;
	}
	return(0);
}

static int eBinWriteFunctionCode(fd, script, code)
	int	fd;
	ember	*script;
	e_code	*code;
{
	e_bin	bin;
	int	length,
		valLength;
	char	*ptr;
	e_funct	*curFunct;
	e_stab	*curSym;

	while(code)
	{
		switch(code->arg.type)
		{
			case EMBER_ARG_FUNCTION:	
				curFunct = (e_funct*)code->arg.ptr;
				ptr = curFunct->name;
				break;

			case EMBER_ARG_VARIABLE:	
				curSym = (e_stab*)code->arg.ptr;
				ptr = curSym->name;
				break;

			default:
				ptr = NULL;
		}
		if (ptr)
			valLength = strlen(ptr) + 1;
		else
			valLength = 0;
		bin.type = EMBER_BIN_CODE;
		length = 4 + 
			4 + 
			4 + 
			4 +
			valLength;
		intToBuf(bin.length,length);
		write(fd, &bin, sizeof(bin));

		writeInt(fd, code->op);
		writeInt(fd, code->line);
		writeInt(fd, code->arg.type);
		writeInt(fd, code->arg.id);
		write(fd, ptr, valLength);

		code = code->next;
	}
	return(0);
}

/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/



int eWriteLibrary(script, path)
	ember	*script;
	char	*path;
{
	int	fd,
		count;
	e_file	header;
	e_funct	*curFunct;


	/*
	** This must be a library so ensure it doesn't have a main()
	*/
	if (script->main.codePtr != NULL)
	{
		strcpy(script->errorMsg,
			"Libraries cannot include a main function!");
		return(-1);
	}

	/*
	** Create the library file
	*/
	fd = open(path, O_RDWR | O_CREAT | O_TRUNC, 0744);
	if (fd < 0)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Couldn't create library file '%s'",
			path);
		return(-1);
	}
	strcpy(header.magic, EMBER_BINARY_MAGIC);
	header.fileType = EMBER_BINARY_LIBRARY;
	header.version =  EMBER_BINARY_VERSION;
	write(fd, &header, sizeof(header));


	/*
	** Write the literal values
	*/

	if (eBinWriteLiterals(fd,script,script->literals)< 0)
	{
		return(-1);
	}

	/*
	** Write the global variables
	*/
	if (eBinWriteGlobalVariables(fd,script)< 0)
	{
		return(-1);
	}
	/*
	** Go through the functions
	*/
        count = 0;
        while(count < EMBER_HASH_LEN)
        {
                curFunct = script->scriptFuncts[count];
                while(curFunct)
                {
			eBinWriteFunctionHeader(fd,curFunct);
			eBinWriteFunctionArgs(fd,curFunct);
                        eBinWriteFunctionVariables(fd,script, 
				curFunct->variables);
                        eBinWriteFunctionCode(fd, script, curFunct->codePtr);
                        curFunct = curFunct->next;
                }
                count++;
        }


	return(0);
}


int eLoadLibrary(script, file)
	ember	*script;
	char	*file;
{
	int	fd,
		res,
		hash,
		binLength,
		bufLength,
		valLength,
		variableOffset,
		literalOffset,
		functionOffset,
		idBuf,
		typeBuf,
		ignore,
		numArgs,
		label;
	char	*cp,
		*valBuf,
		nameBuf[EMBER_NAME_LEN + 1];
	e_file	header;
	e_bin	bin;
	e_stab	*newSym,
		*curSym;
	e_type	*typePtr;
	e_funct	*newFunct,
		*curFunct;
	e_arg	*newArg,
		*curArg;
	e_code	*newCode,
		*curCode,
		*prevCode,
		*tmpCode;

	/*
	** Open the library file
	*/
	if (!file)
	{
		strcpy(script->errorMsg, "Invalid library file name");
		return(-1);
	}
	cp = file + 1;
	*(cp + strlen(cp) - 1) = 0;
	fd = open(cp,O_RDONLY);
	if (fd < 0)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Couldn't open library file '%s'", cp);
		return(-1);
	}

	/*
	** Ensure that it's a library file and that the file format
	** version is correct
	*/
	read(fd, &header, sizeof(header));
	if (strcmp(header.magic, EMBER_BINARY_MAGIC) != 0)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"File '%s' is not an Ember Binary / Library file", cp);
		return(-1);
	}
	if (header.fileType != EMBER_BINARY_LIBRARY)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"File '%s' is not an Ember Library file", cp);
		return(-1);
	}
	if (header.version !=  EMBER_BINARY_VERSION)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Library file '%s' is the wrong version.  Recompile", 
			cp);
		return(-1);
	}

	/*
	** OK, we've got ourselves a library file.  Lets load up the
	** details.  First, we'll find the literals
	*/
	script->info.lastFileID++;
	script->source.names[script->info.lastFileID] = strdup(cp);
	bufLength = 1024;
	valBuf = (char *)eMemMalloc(script, bufLength);
	literalOffset = script->info.lastLitID;
	variableOffset = script->info.lastSymID;
	functionOffset = script->info.lastFunctID;

	res = read(fd, &bin, sizeof(bin));
	if(res < 0)
	{
		strcpy(script->errorMsg, "Library file read failed");
		return(-1);
	}
	if(res == 0)
	{
		return(0);
	}
	while(bin.type == EMBER_BIN_LITERAL)
	{
		/*
		** Read the details from the library
		*/
		binLength = bufToInt(bin.length);
		valLength = binLength - 8;
		newSym = (e_stab*)eMemMalloc(script, sizeof(e_stab));
		bzero(newSym,sizeof(e_stab));
		readInt(fd, &newSym->id);
		readInt(fd, &newSym->value.type);
		if (valLength > 0)
		{
			if (valLength > bufLength)
			{
				free(valBuf);
				bufLength = valLength;
				valBuf=(char *)eMemMalloc(script, bufLength);
			}
			read(fd, valBuf, valLength);
			newSym->value.initialised = EMBER_TRUE;
		}
		else
		{	
			/* This should never be the case */
			newSym->value.initialised = EMBER_FALSE;
		}

		/*
		** Setup the value properly
		*/
		typePtr = eTypeGetType(newSym->value.type);	
		if (!typePtr)
		{
			free(valBuf);
			strcpy(script->errorMsg, "Invalid type in library");
			return(-1);
		}
		(*typePtr->pack)(valBuf, &newSym->value);
		newSym->value.typePtr = typePtr;

		/*
		** Add the literal to the VM
		*/
		newSym->id += literalOffset;
		hash = newSym->id % EMBER_HASH_LEN;
		curSym = script->literals[hash];
		if (curSym == NULL)
		{
			script->literals[hash] = newSym;
		}
		else
		{
			while(curSym->next)
				curSym = curSym->next;
			curSym->next = newSym;
		}
		script->info.lastLitID++;


		res = read(fd, &bin, sizeof(bin));
		if (res < 0)
		{
			free(valBuf);
			strcpy(script->errorMsg, "Library file read failed");
			return(-1);
		}
		if (res == 0)
		{
			free(valBuf);
			return(0);
		}
	}

	/*
	** Read any globals defined in the library
	*/
	while(bin.type == EMBER_BIN_GLOBAL)
	{
		binLength = bufToInt(bin.length);
		valLength = binLength - 8;
		readInt(fd, &idBuf);
		readInt(fd, &typeBuf);
		read(fd, nameBuf, valLength);

		/*
		**  If it doesn't already exist then add it.
		*/
		if (eSymGetVariableByName(script, valBuf) == NULL)
		{
			newSym = (e_stab*)malloc(sizeof(e_stab));
			bzero(newSym, sizeof(e_stab));
			newSym->id = idBuf;
			newSym->value.type = typeBuf;
			strcpy(newSym->name, nameBuf);
			newSym->id += variableOffset;

			hash = newSym->id % EMBER_HASH_LEN;
			curSym = script->variables[hash];
			if (curSym == NULL)
			{
				script->variables[hash] = newSym;
			}
			else
			{
				while(curSym->next)
					curSym = curSym->next;
				curSym->next = newSym;
			}
			eSymAddVariableMap(script, newSym->name, 
				newSym->id);
			script->info.lastSymID++;
		}
		res = read(fd, &bin, sizeof(bin));
		if (res <= 0)
		{
			free(valBuf);
			strcpy(script->errorMsg, 
				"Library file read failed");
			return(-1);
		}
	}

	/*
	** Next we'll find the function definitions
	*/

	while(bin.type == EMBER_BIN_FUNCT)
	{
		/*
		** Read the function header info
		*/
		binLength = bufToInt(bin.length);
		valLength = binLength - 4;
		newFunct = (e_funct *)eMemMalloc(script,sizeof(e_funct));
		bzero(newFunct,sizeof(e_funct));
		readInt(fd,&newFunct->id);
		read(fd,newFunct->name, valLength);
		newFunct->type = EMBER_SCRIPT_FUNCTION;

		/*
		** Is this a duplcate function name
		*/
		curFunct = eSymGetFunctionByName(script,newFunct->name);
		if (curFunct && curFunct->type == EMBER_SCRIPT_FUNCTION)
		{
			free(valBuf);
			snprintf(script->errorMsg, EMBER_ERR_LEN,
				"Library function '%s' is a duplicate name",
				newFunct->name);
			return(-1);
		}

		/*
		** Add the function to the VM
		*/
		script->info.lastFunctID++;
		newFunct->id += functionOffset;
		hash = newFunct->id % EMBER_HASH_LEN;
		curFunct = script->scriptFuncts[hash];
		if (curFunct == NULL)
		{
			script->scriptFuncts[hash] = newFunct;
		}
		else
		{
			while(curFunct->next)
				curFunct = curFunct->next;
			curFunct->next = newFunct;
		}
		script->info.curFunction = newFunct;

		/*
		** Add any function args
		*/
		
		res = read(fd, &bin, sizeof(bin));
		if (res <= 0)
		{
			free(valBuf);
			strcpy(script->errorMsg, "Library file read failed");
			return(-1);
		}
		curArg = NULL;
		numArgs = 0;
		
		while(bin.type == EMBER_BIN_ARG)
		{
			binLength = bufToInt(bin.length);
			valLength = binLength - 8;
			newSym = (e_stab*)eMemMalloc(script,sizeof(e_stab));
			bzero(newSym, sizeof(e_stab));
			readInt(fd, &newSym->id);
			readInt(fd, &newSym->value.type);
			read(fd, newSym->name, valLength);
			newSym->id += variableOffset;
			newSym->value.typePtr=eTypeGetType(newSym->value.type);

			numArgs++;
			hash = newSym->id % EMBER_HASH_LEN;
			curSym = script->variables[hash];
			if (curSym == NULL)
			{
				script->variables[hash] = newSym;
			}
			else
			{
				while(curSym->next)
					curSym = curSym->next;
				curSym->next = newSym;
			}
			eSymAddVariableMap(script, newSym->name, newSym->id);
			script->info.lastSymID++;

			newArg = (e_arg*)eMemMalloc(script,sizeof(e_arg));
			newArg->id = newSym->id;
			strcpy(newArg->name, newSym->name);
			newArg->type = newSym->value.type;
			newArg->ptr = newSym;
			newArg->next = NULL;
			if (curArg == NULL)
			{
				script->info.curFunction->args = newArg;
			}
			else
			{
				curArg->next = newArg;
			}
			curArg = newArg;

			res = read(fd, &bin, sizeof(bin));
			if (res < 0)
			{
				free(valBuf);
				strcpy(script->errorMsg, 
					"Library file read failed");
				return(-1);
			}
		}
		script->info.curFunction->numArgs = numArgs;

		/*
		** Add the function variables
		*/
		while(bin.type == EMBER_BIN_VARIABLE)
		{
			binLength = bufToInt(bin.length);
			valLength = binLength - 8;
			readInt(fd, &idBuf);
			readInt(fd, &typeBuf);
			read(fd, nameBuf, valLength);

			/*
			** Function args are in the variable map already.
			*/
			ignore = 0;
			curArg = script->info.curFunction->args;
			while(curArg)
			{
				if (strcmp(curArg->name, nameBuf) == 0)
				{
					ignore = 1;
					break;
				}
				curArg = curArg->next;
			}

			/*
			**  If it's not an arg then add it
			*/
			if ( ! ignore)
			{
				newSym = (e_stab*)malloc(sizeof(e_stab));
				bzero(newSym, sizeof(e_stab));
				newSym->id = idBuf;
				newSym->value.type = typeBuf;
				strcpy(newSym->name, nameBuf);
				newSym->id += variableOffset;

				hash = newSym->id % EMBER_HASH_LEN;
				curSym = script->variables[hash];
				if (curSym == NULL)
				{
					script->variables[hash] = newSym;
				}
				else
				{
					while(curSym->next)
						curSym = curSym->next;
					curSym->next = newSym;
				}
				eSymAddVariableMap(script, newSym->name, 
					newSym->id);
				script->info.lastSymID++;
			}
			res = read(fd, &bin, sizeof(bin));
			if (res <= 0)
			{
				free(valBuf);
				strcpy(script->errorMsg, 
					"Library file read failed");
				return(-1);
			}
		}

		/*
		** Add the function code
		*/
		curCode = prevCode = NULL;
		while(bin.type == EMBER_BIN_CODE)
		{
			binLength = bufToInt(bin.length);
			valLength = binLength - 16;

			newCode = (e_code*)eMemMalloc(script,sizeof(e_code));
			readInt(fd, &newCode->op);
			readInt(fd, &newCode->line);
			readInt(fd, &newCode->arg.type);
			readInt(fd, &newCode->arg.id);
			read(fd, valBuf,valLength);
			newCode->file = script->info.lastFileID;
			newCode->next = NULL;
			newCode->prev = prevCode;
			prevCode = newCode;
			if(curCode == NULL)
				script->info.curFunction->codePtr = newCode;
			else
				curCode->next = newCode;
			script->info.curFunction->lastPcode = newCode;
			curCode = newCode;

			switch(newCode->arg.type)
			{
			    case EMBER_ARG_VARIABLE:
				newCode->arg.id += variableOffset;
				newCode->arg.ptr=eSymGetVariableByName(script, 
					valBuf);
				if (newCode->arg.ptr == NULL)
				{
				    free(valBuf);
				    sprintf(script->errorMsg,
					"Undefined lib variable %s at line %d",
					valBuf, newCode->line);
				    return(-1);
				}
				break;

			    case EMBER_ARG_LITERAL:
				newCode->arg.id += literalOffset;
				newCode->arg.ptr=eSymGetLiteral(script, 
					newCode->arg.id);
				break;

			    case EMBER_ARG_FUNCTION:
				newCode->arg.id += functionOffset;
				newCode->arg.ptr = eSymGetFunctionByName(script,
					valBuf);
				if (newCode->arg.ptr == NULL)
				{
				    free(valBuf);
				    sprintf(script->errorMsg,
					"Call to unknown function %s at line %d"
					, valBuf, newCode->line);
				    return(-1);
				}
				break;

			    case EMBER_ARG_TYPE:
				newCode->arg.ptr=eTypeGetType(newCode->arg.id);
				break;

			    case EMBER_ARG_LABEL:
				/*
				** We need to sort out the code ptr later
				*/
				break;

			    case EMBER_ARG_ALU:
			    case EMBER_ARG_COND:
				/* 	
				** Don't need to worry about these as
				** there is no pointer to value to
				** restore
				*/
				break;

			}
		
			res = read(fd, &bin, sizeof(bin));
			if (res <= 0)
				break;
		}

		/*
		** Scan the function and repoint any jump destinations
		*/
		curCode = script->info.curFunction->codePtr;
		while(curCode)
		{
			if (curCode->op == E_JMP || curCode->op == E_JZ)
			{
				/* Forward jump */
				label = curCode->arg.id;
				tmpCode = curCode->next;
				while(tmpCode)
				{
					if (tmpCode->op == E_LABEL &&
						tmpCode->arg.id == label)
					{
						curCode->arg.ptr = tmpCode;
						break;
					}
					tmpCode = tmpCode->next;
				}
			}
			if (curCode->op == E_JMPB)
			{
				/* Backwards jump */
				label = curCode->arg.id;
				tmpCode = curCode->prev;
				while(tmpCode)
				{
					if (tmpCode->op == E_LABEL &&
						tmpCode->arg.id == label)
					{
						curCode->arg.ptr = tmpCode;
						break;
					}
					tmpCode = tmpCode->prev;
				}
			}
			curCode = curCode->next;
		}


		/*
		** Anything more to do?
		*/
		if (res < 0)
		{
			free(valBuf);
			strcpy(script->errorMsg, 
				"Library file read failed");
			return(-1);
		}
		if (res == 0)
		{
			free(valBuf);
			return(0);
		}
	}
	free(valBuf);
	return(0);
}
