/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: mod_io.c,v 1.6 2001/10/05 03:25:13 bambi Exp $
**
*/

/*
** Module	: IO Library Module
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/stat.h>

#include <config/config.h>

#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif

#include "ember.h"
#include "parser.h"


/*
** Local type def for a file descriptor object
*/
typedef struct {
        int     fd,
                type,
                mark,
                len;
        char    buf[4096];
        FILE    *fp;  
} fdtab_t;


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/

#define FMT_EMBER_TYPE_ERR "Bad value type for field %d in printf"

#define	FILE_NORMAL	1
#define	FILE_PIPE	2
#define	FILE_FIFO	3

#define	MOD_NAME	"mod_std"
#define	FD_DESCR	"fd"

#ifndef O_BINARY
#  define O_BINARY	0
#endif


static	char	errBuf[EMBER_ERR_LEN + 1];

/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/


static char *_unpackFrame(frame)
	e_stack	*frame;
{
        return((*frame->value.typePtr->unpack)(&frame->value));
}



static char *_formatPrintf(script, args)
	ember	*script;
	e_stack	*args;
{
	e_stack	*curFrame;
	e_type	*intType,
		*uintType,
		*realType;
	char	*fmt,
		*icp,
		*ocp,
		rJust,
		zeroPad,
		*tbp,
		tmpBuf[10 * 1024],
		fmtBuf[24],
		*cp;
	static	char	buf[100 * 1024];
	int	fieldWidth,
		dataWidth,
		widthDiff,
		precision,
		fieldCount = 0;

	/*
	** Get some type pointers we'll need
	*/
	intType = eTypeGetType(EMBER_TYPE_INT);
	uintType = eTypeGetType(EMBER_TYPE_UINT);
	realType = eTypeGetType(EMBER_TYPE_REAL);


	/*
	**
	*/
	(void)bzero(buf,sizeof(buf));
	curFrame = args;
	fmt = curFrame->value.value.textVal;
	if (!fmt)
		return(NULL);

	curFrame = curFrame->next;
	icp = fmt;
	ocp = buf;

	while(*icp)
	{
		switch(*icp)
		{
			case '%':
				icp++;
				rJust = 1;
				if (*icp == '-')
				{
					rJust = 0;
					icp++;
				}
				zeroPad=0;
				if (*icp == '0')
				{
					zeroPad = 1;
					icp++;
				}
	
				fieldWidth = 0;
				dataWidth = 0;
				precision = -1;
				while(isdigit((int)*icp))
				{
					fieldWidth *= 10;
					fieldWidth += *icp - '0';
					icp++;
				}

				if (*icp == '.')
				{
					icp++;
					precision = 0;
					while(isdigit((int)*icp))
					{
						precision *= 10;
						precision += *icp - '0';
						icp++;
					}
				}

				bzero(tmpBuf,sizeof(tmpBuf));
				tbp = tmpBuf;
				switch(*icp)
				{
					char	*tcp;

					case 's':
						fieldCount++;
						if (!curFrame)
						{
							*tbp++ = 's';
							break;
						}
						tcp = _unpackFrame(curFrame);
						strcpy(tbp,tcp);
						curFrame = curFrame->next;
						break;

					case 'd':
						fieldCount++;
						if (!curFrame)
						{
							*tbp++ = 'd';
							break;
						}

						/*
						** Cast to int
						*/
						tcp = _unpackFrame(curFrame);
						(*intType->pack)
							(tcp,&curFrame->value);
						curFrame->value.typePtr = 
							intType;
						tcp = _unpackFrame(curFrame);
						cp = (char *)index(tcp,'.');
						if (cp)
							*cp = 0;
						strcpy(tbp,tcp);
						curFrame = curFrame->next;
						break;

					case 'u':
						fieldCount++;
						if (!curFrame)
						{
							*tbp++ = 'u';
							break;
						}

						/*
						** Cast to unsigned int
						*/
						tcp = _unpackFrame(curFrame);
						(*uintType->pack)
							(tcp,&curFrame->value);
						curFrame->value.typePtr = 
							uintType;
						tcp = _unpackFrame(curFrame);
						strcpy(tbp,tcp);
						curFrame = curFrame->next;
						break;

					case 'f':
						fieldCount++;
						if (precision < 0)
							precision = 6;
						if (!curFrame)
						{
							*tbp++ = 'f';
							break;
						}

						/*
						** Cast to real
						*/
						tcp = _unpackFrame(curFrame);
						(*realType->pack)
							(tcp,&curFrame->value);
						curFrame->value.typePtr = 
							realType;
						
						sprintf(fmtBuf,"%%.%df",
							precision);
						sprintf(tbp,fmtBuf,
						 curFrame->value.value.realVal);
						tbp = tbp + strlen(tbp);
						curFrame=curFrame->next;
						break;

					default:
						*tbp++ = *icp;
				}
				
				dataWidth = strlen(tmpBuf);
				widthDiff = fieldWidth - dataWidth;
				if (widthDiff < 1)
				{
					strcat(ocp,tmpBuf);
					while(*ocp)
						ocp++;
					break;
				}

				if(rJust || zeroPad)
				{
					while(widthDiff)
					{
						if (zeroPad)
							*ocp++ = '0';
						else
							*ocp++ = ' ';
						widthDiff--;
					}
				}
				strcat(ocp,tmpBuf);
				while(*ocp)
					ocp++;
				if(!rJust)
				{
					while(widthDiff)
					{
						*ocp++ = ' ';
						widthDiff--;
					}
				}
				break;

			default:
				*ocp++ = *icp;
		}
		icp++;

	}
	return(buf);
}



static int _addFdtabEntry(script, fd, fp, type)
	ember	*script;
	int	fd;
	FILE	*fp;
	int	type;
{
	fdtab_t	*new;
	int	id;

        new = (fdtab_t *)malloc(sizeof(fdtab_t));
        new->type = type;
        new->fp = fp;
	if (fp)
        	new->fd = fileno(fp);
	else
        	new->fd = fd;
        new->mark = new->len = 0;
	id = eStoreObject(script, MOD_NAME, FD_DESCR, (void *)new);
	if (id < 0)
		free(new);
	return(id);
}

static fdtab_t *_getFdtabEntry(script,id)
	ember	*script;
	int	id;
{
	fdtab_t	*obj;

	obj = (fdtab_t*)eFetchObject(script,MOD_NAME,FD_DESCR,id);
	if (obj == NULL)
		sprintf(script->errorMsg,"Invalid file descriptor (%d)", id);
	return(obj);
}




static char *_bufRead(script,tab)
	ember	*script;
	fdtab_t	*tab;
{
	static	char	cb;

	if (tab->mark == tab->len)
	{
		tab->len = read(tab->fd,tab->buf,4096);
		if(tab->len < 1)
		{
			tab->mark = tab->len = 0;
			return(NULL);
		}
		tab->mark = 0;
	}
	cb = *(tab->buf + tab->mark++);
	return(&cb);
}


static char *_tokRead(script,fd,tok)
	ember	*script;
	int	fd;
	char	*tok;
{
	int	count,
		bail;
	static	char buf[4096];
	char	*newChar;
	register char	*cp,	
			*tp;
	fdtab_t	*cur;
	
	cur = _getFdtabEntry(script,fd);
	if (cur == NULL)
	{
		sprintf(errBuf,"Bad file descriptor (%d).",fd);
		eRunSetError(script, EMBER_TYPE_TEXT, errBuf);
		return(NULL);
	}

	(void)bzero(buf,sizeof(buf));
	cp = buf;
	count = 0;
	bail = 0;
	while(count < sizeof(buf))
	{
		newChar = _bufRead(script,cur);
		if (newChar)
		{
			count++;
			tp = tok;
			while(*tp)
			{
				if (*newChar == *tp)
				{
					bail = 1;
					break;
				}
				tp++;
			}
			*cp++ = *newChar;
		}
		else
			bail = 1;
		if (bail)
			break;
	}
	if (!count)
	{
		return(NULL);
	}
	else
		return(buf);
}



/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/

int doEcho(script, args)
	ember	*script;
	e_stack	*args;
{
	if (script->www.parseHtml && script->www.headersSent == 0)
	{
		eWwwSendResponse(script, NULL);
		eWwwSendHeaders(script);
	}
        printf("%s",emTextArg(args));
        fflush(stdout);
	return(0);
}



int doPrintf(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*outbuf;

	if (script->www.parseHtml && script->www.headersSent == 0)
	{
		eWwwSendResponse(script, NULL);
		eWwwSendHeaders(script);
	}
	outbuf = _formatPrintf(script, args);
	(void)write(fileno(stdout),outbuf,strlen(outbuf));
	return(0);
}


int doFprintf(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*outbuf;
	int	objId;
	fdtab_t	*fd;

	objId = args->value.value.intVal;
	fd = _getFdtabEntry(script,objId);
	if (!fd)
		return(-1);
	outbuf = _formatPrintf(script, args->next);
	(void)write(fd->fd,outbuf,strlen(outbuf));
	return(0);
}

int doSprintf(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*outbuf;

	outbuf = _formatPrintf(script, args);
	eSymSetTextValue(script,&script->returnValue, outbuf);
	return(0);
}



int doOpen(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*path,
		*mode;
	int	fd,
		fileType,
		objId;
	FILE	*pfile,
		*fp = NULL;

	path = emTextArg(args);
	mode = emTextArg(args->next);

	switch(*mode)
	{
		case '>':
			switch(*(mode+1))
			{
				case 'P':	/* Write named pipe */
					fd = mknod(path,S_IFIFO | 0777, 0);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fd=open(path,O_WRONLY|O_CREAT|O_BINARY,
						 0777);
					fileType = FILE_FIFO;
					break;

				case '\0':	/* Write mode */
					fd = open(path,O_WRONLY|O_CREAT|O_TRUNC
						| O_BINARY, 0777);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_NORMAL;
					break;
				case '>': 	/* Append mode */
					fd = open(path,O_WRONLY | O_APPEND
						| O_CREAT | O_BINARY, 0777);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_NORMAL;
					break;
				case '<':	/* Read write mode */
					fd = open(path,O_RDWR|O_CREAT|O_BINARY,
						0777);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_NORMAL;
					break;
				case '|':	/* Write pipe mode */ 
					pfile = popen(path,"w");
					if (pfile)
						fd = fileno(pfile);
					else
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_PIPE;
					fp = pfile;
					break;
				default:
					sprintf(script->errorMsg,
						"Bad file open mode '%s'"
						, mode);
					return(-1);
					break;
			}
			break;
		case '<':
			switch(*(mode+1))
			{
				case 'P':	/* Write named pipe */
					fd = mknod(path,S_IFIFO, 0);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fd = open(path,O_RDONLY);
					fileType = FILE_FIFO;
					break;

				case '\0':	/* Read mode */
					fd = open(path,O_RDONLY|O_BINARY);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_NORMAL;
					break;
				case '|': 	/* Read pipe mode */
					pfile = popen(path,"r");
					if (pfile)
						fd = fileno(pfile);
					else
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
						return(0);
					}
					fileType = FILE_PIPE;
					fp = pfile;
					break;
				case '>':	/* Read write mode */
					fd = open(path,O_RDWR|O_CREAT|O_BINARY,
						0777);
					if (fd < 0)
					{
						eRunSetError(script,EMBER_TYPE_INT,
							strerror(errno));
					}
					fileType = FILE_NORMAL;
					break;
				default:
					sprintf(script->errorMsg,
						"Bad file open mode '%s'"
						, mode);
					return(-1);
					break;
			}
			break;
		default:
			sprintf(script->errorMsg,"Bad file open mode '%s'" , 
				mode);
			return(-1);
			break;
	}
	objId = _addFdtabEntry(script, fd, fp, fileType);
        eSymSetIntValue(script,&script->returnValue, objId);
	return(0);
}



int doClose(script, args)
	ember	*script;
	e_stack	*args;
{
	fdtab_t	*cur;
	int	fd;

	fd = emIntArg(args);
	cur = _getFdtabEntry(script,fd);
	if (cur == NULL)
	{
		sprintf(script->errorMsg,"Bad file descriptor (%d).",fd);
		return(-1);
	}
	switch(cur->type)
	{
		case FILE_NORMAL:
		case FILE_FIFO:
			close(cur->fd);
			break;
		case FILE_PIPE:
			(void)pclose(cur->fp);
			break;
	}
	free(cur);
	return(0);
}



int doReadLn(script, args)
	ember	*script;
	e_stack	*args;
{
	int	fd;
	char	*buf;

	fd = emIntArg(args);

	buf = _tokRead(script, fd,"\n");
	if (!buf)
	{
        	eSymSetTextValue(script,&script->returnValue, "");
	}
	else
	{
        	eSymSetTextValue(script,&script->returnValue, buf);
	}
	return(0);
}
	

void doReadTok(script, args)
	ember	*script;
	e_stack	*args;
{
	int	fd;
	char	*tok,
		*buf;

	tok = emTextArg(args);
	fd = emIntArg(args->next);

	buf = _tokRead(script, fd, tok);
	if (!buf)
	{
        	eSymSetTextValue(script,&script->returnValue, "");
	}
	else
	{
        	eSymSetTextValue(script,&script->returnValue, buf);
	}
}
