/*
**      mod_www.c
**
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <fcntl.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <time.h>

#include "ember.h"
#include "config/config.h"

#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif

static unsigned char isAcceptable[96] =

/* Overencodes */
#define URL_XALPHAS     (unsigned char) 1
#define URL_XPALPHAS    (unsigned char) 2

/*      Bit 0           xalpha          -- see HTFile.h
**      Bit 1           xpalpha         -- as xalpha but with plus.
**      Bit 2 ...       path            -- as xpalpha but with /
*/
    /*   0 1 2 3 4 5 6 7 8 9 A B C D E F */
    {    7,0,0,0,0,0,0,0,0,0,7,0,0,7,7,7,       /* 2x   !"#$%&'()*+,-./ */
         7,7,7,7,7,7,7,7,7,7,0,0,0,0,0,0,       /* 3x  0123456789:;<=>?  */
         7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,       /* 4x  @ABCDEFGHIJKLMNO */
         7,7,7,7,7,7,7,7,7,7,7,0,0,0,0,7,       /* 5X  PQRSTUVWXYZ[\]^_ */
         0,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,       /* 6x  `abcdefghijklmno */
         7,7,7,7,7,7,7,7,7,7,7,0,0,0,0,0 };     /* 7X  pqrstuvwxyz{\}~ DEL */
 
#define ACCEPTABLE(a)   ( a>=32 && a<128 && ((isAcceptable[a-32]) & mask))

static char *hex = "0123456789ABCDEF";


char * HTEscape(str)
        char *str;
{
    unsigned char mask = URL_XPALPHAS;
    char * p;
    char * q;
    char * result;
    int unacceptable = 0;
    for(p=str; *p; p++)
        if (!ACCEPTABLE((unsigned char)*p))
                unacceptable +=2;
    result = (char *) malloc(p-str + unacceptable + 1);
    bzero(result,(p-str + unacceptable + 1));

    if (result == NULL)
    {
	return(NULL);
    }
    for(q=result, p=str; *p; p++) {
        unsigned char a = *p;
        if (!ACCEPTABLE(a)) {
            *q++ = '%';  /* Means hex commming */
            *q++ = hex[a >> 4];
            *q++ = hex[a & 15];
        }
        else *q++ = *p;
    }
    *q++ = 0;                   /* Terminate */
    return result;
}




int doUrlEncode(script, args)
        ember	*script;
	e_stack	*args;
{
        char    *str,
                *new,
                *cp;

        str = emTextArg(args);
        new = (char *)HTEscape(str);
	if (new == NULL)
	{
		eRunSetError(script,EMBER_TYPE_TEXT, "Out of memory");
		return(-1);
	}
        cp = new;
        while(*cp)
        {
                if (*cp == ' ')
                        *cp = '+';
                cp++;
        }
        eSymSetTextValue(script,&script->returnValue, new);
        free(new);
	return(0);
}


int doHttpAuth(script, args)
	ember	*script;
	e_stack	*args;
{
	char	uname[64],
		passwd[64],
		*htmlFile,
		*area;
	e_value	val;

        area = emTextArg(args);
        htmlFile = emTextArg(args->next);
	eWwwAuth(script,uname, passwd, area, htmlFile);

        eSymSetTextValue(script,&val,uname);
        eSymSetArrayElement(script, &script->returnValue, &val, 0);
        eSymSetTextValue(script,&val,passwd);
        eSymSetArrayElement(script, &script->returnValue, &val, 1);
	return(0);

}




int doForceHttpAuth(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*area,
		*file;

        area = emTextArg(args);
        file = emTextArg(args->next);
	if (*file == 0)
		file = NULL;
	eWwwSendAuthHeader(script,area,"Authentication failed", file);
	exit(1);
}


int doAddHttpHeader(script, args)
	ember	*script;
	e_stack	*args;
{
	char 	*header;

	header = emTextArg(args);
	eWwwAddHeader(script, header);
	return(0);
}


int doSetHttpResponse(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*resp;

	resp = emTextArg(args);
	if (script->www.customResponse)
		free(script->www.customResponse);
	script->www.customResponse = strdup(resp);
	return(0);
}

int doSetContentType(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*type;

	type = emTextArg(args);
	if (script->www.contentType)
		free(script->www.contentType);
	script->www.contentType = strdup(type);
	return(0);
}
