/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: runtime.c,v 1.24 2002/06/04 06:37:24 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "config/config.h"
#include "ember.h"
#include "parser.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif
#ifdef HAVE_VARARGS_H
#  include <varargs.h>
#else
#  include <stdarg.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/

#include <ctype.h>


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


#define DEBUG

#ifdef 	DEBUG	
#  define	eDebugTrap()	_debugTrap()
#else
#  define	eDebugTrap()
#endif

#define eDebugTrace(s,x)	if (s->info.traceInfo)printf x ;
#ifdef iswhite
#  undef iswhite
#endif
#define iswhite(x) (x == ' ' || x == '\t' || x == '\n')

static	e_type	*intTypePtr = NULL;

/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

#define	EMBER_MAX_FREE_FRAMES	20

static	e_stack *spareFrames[EMBER_MAX_FREE_FRAMES],
		*lastFrame = NULL;
static	int	curFreeFrames = 0;



static void _freeFrame(script, frame)
	ember	*script;
	e_stack	*frame;
{
	eSymFreeValue(script, &frame->value);
	if (lastFrame == NULL)
	{
		lastFrame = frame;
	}
	else
	{
		if (curFreeFrames < EMBER_MAX_FREE_FRAMES)
		{
			spareFrames[curFreeFrames] = frame;
			curFreeFrames++;
		}
		else
		{
			free(frame);
		}
	}
}


static void _freeArgList(script, argHead)
	ember	*script;
	e_stack	*argHead;
{
	e_stack	*curArg;

	curArg = argHead;
	while(curArg)
	{
		argHead = curArg->next;
		_freeFrame(script,curArg);
		curArg = argHead;
	}
}




void eFreeStackFrames()
{
	int	count;

	count = 0;
	while(count < curFreeFrames)
	{
		free(spareFrames[count]);
		count++;
	}
	curFreeFrames = 0;
}


static e_stack *_newFrame()
{
	e_stack	*new;

	if (lastFrame)
	{
		new = lastFrame;
		lastFrame = NULL;
	}
	else
	{
		if (curFreeFrames > 0)
		{
			curFreeFrames--;
			new = spareFrames[curFreeFrames];
		}
		else
		{
			new = (e_stack *)malloc(sizeof(e_stack));
		}
	}
	new->value.initialised = 0;
	new->value.value.intVal = 0;
	return(new);
}


e_stack *ePopFrame(script)
	ember	*script;
{
	e_stack	*cur;

	if (script->valueStack == NULL)
	{
		strcpy(script->errorMsg,"Stack Underflow");
		return(NULL);
	}
	cur = script->valueStack;
	script->valueStack = cur->next;
	cur->next = NULL;
	return(cur);
}


void ePushFrame(script, frame)
	ember	*script;
	e_stack	*frame;
{
	frame->next = script->valueStack;
	script->valueStack = frame;
	return;
}



static int _pushValue(script, value, isLiteral)
	ember	*script;
	e_value	*value;
	int	isLiteral;
{
	REG 	e_stack	*frame;

	/*
	** Setup our new stack frame
	*/
	frame = _newFrame();
	if (script->valueStack)
	{
		frame->next = script->valueStack;
	}
	else
	{
		frame->next = NULL;
	}
	script->valueStack = frame;

	/*
	** Drop the value onto the frame
	*/
	if (isLiteral)
	{
		if (eSymExpandLiteral(script,value,&(frame->value)) < 0)
			return(-1);
	}
	else
	{
		eSymDupValue(script,value,&(frame->value));
	}
	frame->block = script->info.curBlock;
	return(0);
}



static 	e_fstck	*lastFunction = NULL;

static void _pushFunction(script)
	ember	*script;
{
	e_fstck	*new;

	if (lastFunction == NULL)
	{
		new = (e_fstck *)malloc(sizeof(e_fstck));
	}
	else
	{
		new = lastFunction;
		lastFunction = NULL;
	}
	new->funct = script->info.curFunction;
	new->next = script->functStack;
	script->functStack = new;
}

static void _popFunction(script)
	ember	*script;
{
	e_fstck	*tmp;

	tmp = script->functStack;
	script->info.curFunction = tmp->funct;
	script->functStack = script->functStack->next;
	if (lastFunction == NULL)
		lastFunction = tmp;
	else
		free(tmp);
}

#ifdef DEBUG
void _debugTrap()
{
	/* Just somewhere to set a breakpoint */
}
#endif

/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/

#ifdef HAVE_VARARGS_H
void eRunSetError(va_alist)
	va_dcl
#else
void eRunSetError(script, type, ...)
#endif
{
	va_list	args;
	char	*fmt,
		errorBuf[1024];
	e_stab	*variable;

#ifdef HAVE_VARARGS_H
	ember	*script;
	int	type;

	va_start(args);
	script = (ember*)va_arg(args, ember*);
	type = (int)va_arg(args, int);
#else
	va_start(args, type);
#endif
        fmt = (char *)va_arg(args,char *);
        vsprintf(errorBuf,fmt,args);
        va_end(args);

	variable = eSymGetVariableByName(script, "@ERRMSG");
	if (variable == NULL)
	{
		/* must be during a module initialisation of similar */
		variable = eSymCreateVariable(script, "@ERRMSG");
	}
	eSymSetTextValue(script, &variable->value, errorBuf);
	switch(eTypeGetRepType(type))
	{
		case EMBER_TYPE_INT:
			eSymSetIntValue(script, &script->returnValue, -1);
			break;

		case EMBER_TYPE_UINT:
			eSymSetUintValue(script, &script->returnValue, -1);
			break;

		case EMBER_TYPE_REAL:
			eSymSetRealValue(script, &script->returnValue, -1);
			break;

		case EMBER_TYPE_TEXT:
			eSymSetTextValue(script, &script->returnValue, "");
			break;
	}
}



char *eRunExpandText(script, str)
	ember	*script;
	char	*str;
{
	static	char buf[EMBER_TEXT_LEN];
	char	*inPtr,*outPtr,
		*cp,
		*valPtr = NULL,
		*tokStart,*tokEnd, 
		nameBuf[EMBER_NAME_LEN], elemBuf[80],
		tmpBuf[2];
	e_stab	*sym,
		*deref;
	e_value	*curVal;
	int	element;

	inPtr = str;
	outPtr = buf;
	while (*inPtr)
	{
		switch(*inPtr)
		{
		    case '$':
		    case '@':
			tokStart = inPtr;
			tokEnd = inPtr+1;
			cp = nameBuf;
			*cp++ = *inPtr;
			while(isalnum((int)*tokEnd) || *tokEnd == '_')
			{
				*cp++ = *tokEnd++;
			}
			*cp = 0;
			sym = eSymGetVariableByName(script, nameBuf);

			if (sym != NULL && *tokEnd == '[')
			{
				tokEnd++;
				cp = elemBuf;
				while(*tokEnd && *tokEnd != ']')
					*cp++ = *tokEnd++;
				*cp = 0;

				tokStart = elemBuf;
				while (*tokStart && iswhite((int)*tokStart))
				{
					tokStart++;
				}

				if (*tokEnd != ']')
				{
					sprintf(script->errorMsg,
						"Bad array reference for %s",
						nameBuf);
					return(NULL);
				}
				if(cp > elemBuf)
					cp--;
				while(iswhite((int)*cp))
				{
					*cp = 0;
					cp--;
				}

				if (*elemBuf=='$'||*elemBuf=='@')
				{
					/*
					** The element reference is in 
					** a variable
					*/
					deref = eSymGetVariableByName(script, 
						elemBuf);
					if (!deref)
						return(NULL);
					switch(deref->value.type)
					{
					case EMBER_TYPE_INT:
						element = 
						deref->value.value.intVal;
						break;

					case EMBER_TYPE_UINT:
						element = 
						deref->value.value.uintVal;
						break;

					default:
						sprintf(script->errorMsg,
						"Array offset must be int ($%s)"
						,elemBuf);
						return(NULL);
					}
				}
				else
				{
					/*
					** The element reference is a
					** literal int value!
					*/
					cp = elemBuf;
					while(*cp && isdigit((int)*cp))
						cp++;
					if (*cp)
					{
					    sprintf(script->errorMsg,
						"Bad array reference for %s",
						nameBuf);
					    return(NULL);
					}
					element = atoi(elemBuf);
				}
				if (sym->value.type == EMBER_TYPE_ARRAY)
				{
					curVal = eSymGetArrayElement(script,
						&sym->value, element);
					valPtr = NULL;
					if (curVal && curVal->initialised)
					{
					    valPtr=(*curVal->typePtr->unpack)
						(curVal);
					}
					inPtr = tokEnd + 1;
				}
				if (sym->value.type == EMBER_TYPE_TEXT)
				{
					char 	*cp;

					cp = sym->value.value.textVal;
					if (element > strlen(cp))
					{
						*tmpBuf = 0;
					}
					else
					{
						tmpBuf[0] = *(cp + element);
						tmpBuf[1] = 0;
					}
					valPtr = tmpBuf;
					inPtr = tokEnd + 1;
				}
			}
			else
			{
				if (sym && sym->value.initialised)
					valPtr = emUnpackSymbol(sym);
				else
					valPtr = nameBuf;
				inPtr += strlen(nameBuf);
			}
			if (valPtr)
			{
				strcpy(outPtr, valPtr);
				outPtr += strlen(valPtr);
			}
			break;

		    case '\\':
			inPtr++;
			if (*inPtr)
			{
				switch(*inPtr)
				{
					case 'n':
						*outPtr++ = '\n';
						break;
					case 't':
						*outPtr++ = '\t';
						break;
					case 'r':
						*outPtr++ = '\r';
						break;
					default:
						*outPtr++ = *inPtr;
						break;
				}
				inPtr++;
			}
			else
			{
				*outPtr = '\\';
			}
			break;

		    default:
			*outPtr++ = *inPtr++;
			break;
		}
	}
	*outPtr = 0;
	return(buf);
}



int eRunPush(script, symbol)
	ember	*script;
	e_stab	*symbol;
{
	int	result;

	/*
	** If this symbol has no value then it's an error
	*/
	if (!symbol->value.initialised && symbol->value.type != EMBER_TYPE_ARRAY)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Reference to uninitialised variable %s",
			symbol->name);
		eDebugTrap();
		return(-1);
	}

	if (emIsLiteral(symbol->name))
	{
		result = _pushValue(script, &symbol->value, EMBER_TRUE);
	}
	else
	{
		result = _pushValue(script, &symbol->value, EMBER_FALSE);
	}
	return(result);
}


int eRunPop(script)
	ember	*script;
{
	e_stack	*frame;

	frame = ePopFrame(script);
	if (frame == NULL)
		return(-1);
	_freeFrame(script,frame);
	return(0);
}


int eRunStore(script, symbol)
	ember	*script;
	e_stab	*symbol;
{
	e_stack	*frame;

	frame = ePopFrame(script);
	if (frame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	eSymFreeValue(script,&symbol->value);
	eSymDupValue(script, &frame->value, &symbol->value);
	symbol->value.initialised = EMBER_TRUE;
	_freeFrame(script,frame);
	return(0);
}



int eRunArrayDeref(script)
	ember	*script;
{
	e_stack	*valFrame,
		*derefFrame;
	e_array	*curElement;
	e_value	*value,
		tmpVal;
	int	offset,
		count;
	char	*textRep;
	static	e_stab *blankValue = NULL;

	if (blankValue == NULL)
	{
		int	id;

        	if(eSymCreateLiteral(script,EMBER_TYPE_TEXT,"\"\"",&id) < 0)
                	return(-1);
        	blankValue = eSymGetLiteral(script, id);
	}

	valFrame = ePopFrame(script);
	if (valFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	if (valFrame->value.type != EMBER_TYPE_ARRAY && 
		valFrame->value.type != EMBER_TYPE_TEXT)
	{
		strcpy(script->errorMsg, 
			"Bad type for deref.  Must be ARRAY or TEXT");
		_freeFrame(script,valFrame);
		return(-1);
	}
	derefFrame = ePopFrame(script);
	if (derefFrame == NULL)
	{
		eDebugTrap();
		_freeFrame(script,valFrame);
		return(-1);
	}

        /*
        ** Can we use the deref info?
        */
	if (intTypePtr == NULL)
		intTypePtr = eTypeGetType(EMBER_TYPE_INT);
        value = &derefFrame->value;
        switch(value->type)
        {
                case EMBER_TYPE_INT:
                        offset = value->value.intVal;
                        break;
                case EMBER_TYPE_UINT:
                        offset = value->value.uintVal;
                        break;
                default:
                        textRep = (*value->typePtr->unpack)(value);
                        if ((*intTypePtr->pack)(textRep,value) < 0)
                        {
                                strcpy(script->errorMsg,
                                        "Bad type for array dereference");
				_freeFrame(script,valFrame);
				_freeFrame(script,derefFrame);
                                return(-1);
                        }
                        offset = value->value.intVal;
                        break;
        }


	/*
	** Deref the variable
	*/
	if (valFrame->value.type == EMBER_TYPE_ARRAY)
	{
		curElement = (e_array*)valFrame->value.value.arrayVal;
		count = 0;
		while(curElement && count < offset)
		{
			curElement = curElement->next;
			count++;
		}
		if (curElement == NULL || 
		    curElement->value.initialised == EMBER_FALSE)
		{
			if (_pushValue(script,&blankValue->value,EMBER_FALSE)<0)
			{
				_freeFrame(script,valFrame);
				_freeFrame(script,derefFrame);
				return(-1);
			}
		}
		else
		{
			if(_pushValue(script,&curElement->value,EMBER_FALSE)<0)
			{
				_freeFrame(script,valFrame);
				_freeFrame(script,derefFrame);
				return(-1);
			}
		}
	}

	if (valFrame->value.type == EMBER_TYPE_TEXT)
	{
		char	*cp,
			tmpBuf[2];

		cp = valFrame->value.value.textVal;
		if (strlen(cp) < offset)
		{
			if (_pushValue(script,&blankValue->value,EMBER_FALSE)<0)
			{
				_freeFrame(script,valFrame);
				_freeFrame(script,derefFrame);
				return(-1);
			}
		}
		else
		{

			tmpBuf[0] = *(cp + offset);
			tmpBuf[1] = 0;
			eSymSetTextValue(script, &tmpVal, tmpBuf);
			_pushValue(script,&tmpVal, EMBER_FALSE);
			free(tmpVal.value.textVal);
		}
	}
	_freeFrame(script,valFrame);
	_freeFrame(script,derefFrame);
	return(0);
}


int eRunAssocDeref(script)
	ember	*script;
{
	e_stack	*valFrame,
		*derefFrame;
	e_assoc	*assoc;
	e_value	*value;
	char	*index;
	static	e_stab *blankValue = NULL;

	if (blankValue == NULL)
	{
		int	id;

        	if(eSymCreateLiteral(script,EMBER_TYPE_TEXT,"\"\"",&id) < 0)
                	return(-1);
        	blankValue = eSymGetLiteral(script, id);
	}

	valFrame = ePopFrame(script);
	if (valFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	if (valFrame->value.type != EMBER_TYPE_ASSOC)
	{
		strcpy(script->errorMsg, "Variable not an assoc array");
		_freeFrame(script,valFrame);
		return(-1);
	}
	derefFrame = ePopFrame(script);
	if (derefFrame == NULL)
	{
		eDebugTrap();
		_freeFrame(script,valFrame);
		return(-1);
	}

        /*
        ** Can we use the deref info?
        */
	if (intTypePtr == NULL)
		intTypePtr = eTypeGetType(EMBER_TYPE_INT);
        value = &derefFrame->value;
	if (value->type != EMBER_TYPE_TEXT)
	{
		strcpy(script->errorMsg, "Bad type for array dereference");
		_freeFrame(script,valFrame);
		_freeFrame(script,derefFrame);
		return(-1);
	}
	index = value->value.textVal;


	/*
	** Deref the variable
	*/
	assoc = (e_assoc*)valFrame->value.value.arrayVal;
	while(assoc)
	{
		if (strcmp(assoc->index, index) == 0)
			break;
		assoc = assoc->next;
	}
	if (assoc == NULL)
	{
		if (_pushValue(script,&blankValue->value,EMBER_FALSE)<0)
		{
			_freeFrame(script,valFrame);
			_freeFrame(script,derefFrame);
			return(-1);
		}
	}
	else
	{
		if(_pushValue(script,&assoc->value,EMBER_FALSE)<0)
		{
			_freeFrame(script,valFrame);
			_freeFrame(script,derefFrame);
			return(-1);
		}
	}
	_freeFrame(script,valFrame);
	_freeFrame(script,derefFrame);
	return(0);
}


int eRunCount(script)
	ember	*script;
{
	e_stack	*valFrame;
	e_value	*value,
		result;
	int	count;

	valFrame = ePopFrame(script);
	if (valFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	value = &valFrame->value;

	if (value->type != EMBER_TYPE_ARRAY && 
		value->type != EMBER_TYPE_TEXT &&
		value->type != EMBER_TYPE_ASSOC )
	{
		strcpy(script->errorMsg, "Bad type for count operation");
		_freeFrame(script,valFrame);
		return(-1);
	}
	if (value->type == EMBER_TYPE_ARRAY)
	{
		count = eSymGetNumArrayElements(script, value);
	}
	if (value->type == EMBER_TYPE_ASSOC)
	{
		count = eSymGetNumAssocElements(script, value);
	}
	if (value->type == EMBER_TYPE_TEXT)
	{
		count = strlen(value->value.textVal);
	}
	eSymSetIntValue(script,&result, count);
	_pushValue(script, &result, EMBER_FALSE);
	_freeFrame(script,valFrame);
	return(0);
}



int eRunArrayStore(script, symbol)
	ember	*script;
	e_stab	*symbol;
{
	e_stack	*valFrame,
		*derefFrame;
	e_value	*value;
	int	offset,
		result;
	char	*textRep;

	if (intTypePtr == NULL)
		intTypePtr = eTypeGetType(EMBER_TYPE_INT);

	/*
	** If this thing isn't an array then fix it
	*/
	if (symbol->value.type != EMBER_TYPE_ARRAY)
	{
		eSymFreeValue(script,&symbol->value);
		symbol->value.initialised = EMBER_FALSE;
	}

	/*
	** Get the value and the dereference info
	*/
	valFrame = ePopFrame(script);
	if (valFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	derefFrame = ePopFrame(script);
	if (derefFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}

	/*
	** Can we use the deref info?
	*/
	value = &derefFrame->value;
	switch(value->type)
	{
		case EMBER_TYPE_INT:
			offset = value->value.intVal;
			break;
		case EMBER_TYPE_UINT:
			offset = value->value.uintVal;
			break;
		default:
			textRep = (*value->typePtr->unpack)(value);
			if ((*intTypePtr->pack)(textRep,value) < 0)
			{
				_freeFrame(script,valFrame);
				_freeFrame(script,derefFrame);
				strcpy(script->errorMsg, 
					"Bad type for array dereference");
				return(-1);
			}
			offset = value->value.intVal;
			break;
	}

	result = eSymSetArrayElement(script, &symbol->value, &valFrame->value, 
		offset);
	symbol->value.type = EMBER_TYPE_ARRAY;
	_freeFrame(script,valFrame);
	_freeFrame(script,derefFrame);
	symbol->value.initialised = EMBER_TRUE;
	return(result);
}



int eRunAssocStore(script, symbol)
	ember	*script;
	e_stab	*symbol;
{
	e_stack	*valFrame,
		*derefFrame;
	e_value	*value;
	int	result;
	char	*index;

	/*
	** If this thing isn't an assoc then fix it
	*/
	if (symbol->value.type != EMBER_TYPE_ASSOC)
	{
		eSymFreeValue(script,&symbol->value);
		symbol->value.initialised = EMBER_FALSE;
	}

	/*
	** Get the value and the dereference info
	*/
	valFrame = ePopFrame(script);
	if (valFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}
	derefFrame = ePopFrame(script);
	if (derefFrame == NULL)
	{
		eDebugTrap();
		return(-1);
	}

	/*
	** Can we use the deref info?
	*/
	value = &derefFrame->value;
	if (value->type != EMBER_TYPE_TEXT)
	{
		_freeFrame(script,valFrame);
		_freeFrame(script,derefFrame);
		strcpy(script->errorMsg, 
			"Bad type for assoc dereference");
		return(-1);
	}
	index = value->value.textVal;
	result = eSymSetAssocElement(script, &symbol->value, 
			&valFrame->value, index);
	symbol->value.type = EMBER_TYPE_ASSOC;
	_freeFrame(script,valFrame);
	_freeFrame(script,derefFrame);

	symbol->value.initialised = EMBER_TRUE;
	return(result);
}



int eRunCast(script, typePtr)
	ember	*script;
	e_type	*typePtr;
{
	e_stack	*frame;
	char	*oldVal,
		*textRep;
	int	length;

	frame = ePopFrame(script);
	if (frame == NULL)
		return(-1);
	if (frame->value.initialised == EMBER_FALSE)
	{
		strcpy(script->errorMsg,
			"Type cast of uninitialised variable failed");
		_freeFrame(script,frame);
		return(-1);
	}
	if (eTypeGetBaseType(frame->value.type) == EMBER_TYPE_TEXT)
		oldVal = frame->value.value.textVal;
	else
		oldVal = NULL;
	textRep = (*frame->value.typePtr->unpack)(&frame->value);
	length = (*typePtr->pack)(textRep,&frame->value);
	if (length < 0)
	{
		strcpy(script->errorMsg,"Type cast failed");
		_freeFrame(script,frame);
		return(-1);
	}
	if (oldVal)
		eMemFree(script,oldVal);

	if(_pushValue(script, &frame->value, EMBER_FALSE) < 0)
	{
		_freeFrame(script,frame);
		return(-1);
	}
	_freeFrame(script,frame);
	return(0);
}



int eRunExists(script, sym)
	ember	*script;
	e_stab	*sym;
{
	char	*varName;
	e_stab	*tmp;
	static	e_stab trueSym, falseSym;

	if (trueSym.value.initialised == 0)
	{
		trueSym.value.value.intVal = 1;
		trueSym.value.initialised = EMBER_TRUE;
		trueSym.value.type = EMBER_TYPE_INT;
		trueSym.value.typePtr = eTypeGetType(EMBER_TYPE_INT);

		falseSym.value.initialised = EMBER_TRUE;
		falseSym.value.type = EMBER_TYPE_INT;
		falseSym.value.typePtr = eTypeGetType(EMBER_TYPE_INT);
	}

	varName = sym->value.value.textVal;
	tmp = eSymGetVariableByName(script, varName);
	if (!tmp)
	{
		eRunPush(script, &falseSym);
		return(0);
	}
	if (tmp->value.initialised == EMBER_TRUE)
	{
		eRunPush(script, &trueSym);
	}
	else
	{
		eRunPush(script, &falseSym);
	}
	return(0);
}



int eRunAlu(script, op)
	ember	*script;
	int	op;
{
	e_stack	*frame1,
		*frame2;
	e_value	*value1,
		*value2,
		result;

	/*
	** Pop the 2 frames we need off the stack (they are in reverse
	** order).
	*/
	frame2 = ePopFrame(script);
	if (frame2 == NULL)
		return(-1);
	frame1 = ePopFrame(script);
	if (frame1 == NULL)
	{
		_freeFrame(script,frame2);
		return(-1);
	}
	value1 = & frame1->value;
	value2 = & frame2->value;

	/*
	** Can we perform this operation given the types of the values?
	*/
	if (eTypeCheckMathOp(script, op, value1, value2) < 0)
	{
		_freeFrame(script,frame1);
		_freeFrame(script,frame2);
		return(-1);
	}

	if ((*frame1->value.typePtr->math)(script,value1,
		value2, op, &result) < 0)
	{
		_freeFrame(script,frame1);
		_freeFrame(script,frame2);
		return(-1);
	}
	_freeFrame(script,frame1);
	_freeFrame(script,frame2);
	result.initialised = 1;
	if(_pushValue(script, &result, EMBER_FALSE) < 0)
		return(-1);
	eSymFreeValue(script, &result);
	return(0);
}



int eRunCmp(script, op)
	ember	*script;
	int	op;
{
	e_stack	*frame1,
		*frame2;
	e_value	*value1,
		*value2;
	static	e_value resValue;
	static	int	init;
	int	result;

	/*
	** Pop the 2 frames we need off the stack (they are in reverse
	** order).
	*/
	frame2 = ePopFrame(script);
	if (frame2 == NULL)
		return(-1);
	frame1 = ePopFrame(script);
	if (frame1 == NULL)
	{
		_freeFrame(script,frame2);
		return(-1);
	}
	value1 = & frame1->value;
	value2 = & frame2->value;

	/*
	** Can we perform this operation given the types of the values?
	*/
	if (eTypeCheckMathOp(script, op, value1, value2) < 0)
	{
		_freeFrame(script,frame1);
		_freeFrame(script,frame2);
		return(-1);
	}

	if ((*frame1->value.typePtr->compare)(script, value1, value2,
		op, &result) < 0)
	{
		_freeFrame(script,frame1);
		_freeFrame(script,frame2);
		return(-1);
	}
	_freeFrame(script,frame1);
	_freeFrame(script,frame2);

	if (init == 0)
	{
		resValue.length = 4;
		resValue.initialised = EMBER_TRUE;
		resValue.type = EMBER_TYPE_INT;
		resValue.typePtr = eTypeGetType(EMBER_TYPE_INT);
		init = 1;
	}
	resValue.value.intVal = result;

	if(_pushValue(script, &resValue, EMBER_FALSE) < 0)
		return(-1);
	return(0);
}


int eRunJz(script, isZero)
	ember	*script;
	int	*isZero;
{
	e_stack	*frame;
	e_value	*value;

	frame = ePopFrame(script);
	if (frame == NULL)
		return(-1);
	value = & frame->value;
	if (value->type != EMBER_TYPE_INT)
	{
		_freeFrame(script,frame);
		strcpy(script->errorMsg,"JZ comparison to non int!");
		return(-1);
	}
	if (value->value.intVal == 0)
		*isZero = 1;
	else
		*isZero = 0;
	_freeFrame(script,frame);
	return(0);
}


int eRunCall(script, funct)
	ember	*script;
	e_funct *funct;
{
	int	result,
		count,
		curBlock,
		argCount,
		numArgs;
	e_stack	*curFrame,
		*argHead,
		*curArg;
	e_arg	*argDef;


	/*
	** Create a param list to hand to the function.  Note
	** we want this off the stack and in another list so
	** that we can free it when the function is done.  We
	** do not want the function to be responsible for freeing
	** stack frames!  This limits the risk of memory leaks in
	** external module routines
	*/

	argCount = 0;
	argHead = NULL;
	curBlock = script->info.curBlock;
	curFrame = script->valueStack;
	while(curFrame && curFrame->block == curBlock)
	{
		argCount++;
		curArg = curFrame;
		curFrame = curFrame->next;
		curArg->next = argHead;
		argHead = curArg;
	}
	script->valueStack = curFrame;
	script->returnValue.initialised = 0;

	/*
	** Execute the function.  
	*/
	switch(funct->type)
	{
	    case EMBER_SCRIPT_FUNCTION:

		/*
		** Check out the args
		*/
		numArgs = funct->numArgs;
		if (numArgs != argCount)
		{
			snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Bad arg count for call to %s. Expected %d, got %d",
				funct->name, numArgs, argCount);
			_freeArgList(script,argHead);
			return(-1);
		}

		count = 0;
		curArg = argHead;
		argDef = funct->args;
		while(count < numArgs && curArg)
		{
			if (curArg->value.type != argDef->type)
			{
				snprintf(script->errorMsg, EMBER_ERR_LEN,
				"Bad type for arg %d in call to %s.",
				count + 1, funct->name);
				_freeArgList(script,argHead);
				return(-1);
			}

			eSymFreeValue(script,&argDef->ptr->value);
			eSymDupValue(script, &curArg->value,
				&argDef->ptr->value);
			curArg->value.initialised = EMBER_TRUE;

			curArg = curArg->next;
			argDef = argDef->next;
			count++;
		}
		_freeArgList(script,argHead);
		
		/*
		** Run the script pcode.  Note the there will be a value
		** left on the stack (either a real value or just an
		** empty forced on there by the parser).  Put that value
		** into the callers execution block before we bail
		*/
		if(eRunCode(script, funct) < 0)
			return(-1);
		script->valueStack->block--;
		return(0);
		break;

	    case EMBER_MODULE_FUNCTION:

		/*
		** If it's a vararg function then check as many of the
		** args as we can and then punt the rest of the checking
		** at the external function.  There's only so much we can
		** do.
		*/
	
		numArgs  = funct->externPtr->numArgs;
		if (numArgs == -1)
		{
			result = (*funct->externPtr->functPtr)(script, argHead);
			if(script->returnValue.initialised == 0)
			{
				eSymSetIntValue(script,&script->returnValue,0);
			}
			script->info.curBlock--;
			if(_pushValue(script,&script->returnValue,
				EMBER_FALSE)< 0)
			{
				_freeArgList(script,argHead);
				return(-1);
			}
			script->info.curBlock++;
			_freeArgList(script,argHead);
			return(result);
		}

		/*
		** It's not a varargs function so step through the args
		** and check the type.  Remember that the args are on
		** the stack in reverse order
		*/
		if (numArgs != argCount)
		{
			snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Bad arg count for call to %s. Expected %d, got %d",
				funct->name, numArgs, argCount);
			_freeArgList(script,argHead);
			return(-1);
		}

		count = 0;
		curArg = argHead;
		while(count < numArgs && curArg)
		{
			if (curArg->value.type != 
				funct->externPtr->argTypes[count])
			{
				snprintf(script->errorMsg, EMBER_ERR_LEN,
				"Bad type for arg %d in call to %s.",
				count + 1, funct->name);
				_freeArgList(script,argHead);
				return(-1);
			}
			curArg = curArg->next;
			count++;
		}
		
		result = (*funct->externPtr->functPtr)(script, argHead);
		_freeArgList(script,argHead);

		/*
		** Check out the function return.  There must ALWAYS be a
		** result frame pushed onto the stack.  If the function
		** doesn't set a return value then we just push a zero
		** value.  There MUST be something left on the stack!  
		** Make sure the result frame is in the execution block 
		** of the caller, not the function.
		*/
		if(script->returnValue.initialised == 0)
		{
			eSymSetIntValue(script,&script->returnValue,0);
		}
		script->info.curBlock--;
		if(_pushValue(script, &script->returnValue, EMBER_FALSE) < 0)
		{
			return(-1);
		}
		eSymFreeValue(script, &script->returnValue);
		script->info.curBlock++;
		break;
	}
	return(result);
}


int eRunCode(script, funct)
	ember	*script;
	e_funct	*funct;
{
	REG	e_code	*curCode;
	int	isZero;

	curCode = funct->codePtr;
	script->info.curFunction = funct;
	*script->errorMsg = 0;
	while(curCode)
	{
		if (script->exitNow == EMBER_TRUE)
			return(0);
		script->info.curLineNum = curCode->line;
		script->info.curFileID = curCode->file;
		switch(curCode->op)
		{
			case E_PUSH:
				if (eRunPush(script, curCode->arg.ptr) < 0)
					return(-1);
				break;

			case E_POP:
				if (eRunPop(script) < 0)
					return(-1);
				break;

			case E_STOR:
				if (eRunStore(script, curCode->arg.ptr) < 0)
					return(-1);
				break;

			case E_ARRAY_STOR:
				if (eRunArrayStore(script, curCode->arg.ptr)<0)
					return(-1);
				break;

			case E_ASSOC_STOR:
				if (eRunAssocStore(script, curCode->arg.ptr)<0)
					return(-1);
				break;

			case E_CAST:
				if (eRunCast(script, curCode->arg.ptr) < 0)
					return(-1);
				break;

			case E_ALU:
				if (eRunAlu(script, curCode->arg.id) < 0)
					return(-1);
				break;

			case E_CMP:
				if (eRunCmp(script, curCode->arg.id) < 0)
					return(-1);
				break;

			case E_JZ:
				if (eRunJz(script, &isZero) < 0)
					return(-1);
				if (isZero == EMBER_FALSE)
					break;
				/* fall through and take the jump */

			case E_JMP:
			case E_JMPB:
				if (curCode->arg.ptr == NULL)
				{
					strcpy(script->errorMsg,
						"Missing jump pointer!");
					return(-1);
				}
				curCode = (e_code *)curCode->arg.ptr;
				break;

			case E_CALL:
				_pushFunction(script);
				if (curCode->arg.ptr == NULL)
				{
					curCode->arg.ptr=eSymGetFunctionById(
						script, curCode->arg.id);
				}
				if (eRunCall(script, curCode->arg.ptr) < 0)
				{
					_popFunction(script);
					return(-1);
				}
				_popFunction(script);
				*script->errorMsg = 0;
				break;

			case E_ARRAY_DEREF:
				if (eRunArrayDeref(script) < 0)
					return(-1);
				break;

			case E_ASSOC_DEREF:
				if (eRunAssocDeref(script) < 0)
					return(-1);
				break;

			case E_COUNT:
				if (eRunCount(script) < 0)
					return(-1);
				break;

			case E_EXISTS:
				if (eRunExists(script, curCode->arg.ptr) < 0)
					return(-1);
				break;

			case E_BLOCK_IN:
				script->info.curBlock++;
				break;

			case E_BLOCK_OUT:
				script->info.curBlock--;
				break;

			case E_RETURN:
				return(0);
				break;

		}
		curCode = curCode->next;
	}
	return(0);
}




#ifdef NOTDEF

int eCallFunction(script, funct, numArgs, args)
	ember	*script;
	char	*funct;
	int	numArgs;
	e_stack	args[];
{
	ember	tmpScript;
	e_funct	*functPtr;
	e_stack	*curArg,
		*prevArg;
	int	count,
		result;

	count = 0;
	prevArg = NULL;
	while(count < numArgs)
	{
		curArg = &args[count];
		if (prevArg)
		{
			prevArg->next = curArg;
		}
		curArg->next = NULL;
		prevArg = curArg;
		count++;
	}
	functPtr = eSymGetFunctionByName(script, funct);
	if (functPtr == NULL)
		return(-1);

	result = (*functPtr->externPtr->functPtr)(&tmpScript, args);
	return(0);
}

#endif



int eRunScript(script)
	ember	*script;
{
	/*
	** Run the script.  Quite simply, just jump in and execute the
	** main function if it exists
	*/

	if (script->main.codePtr == NULL)
	{
		strcpy(script->errorMsg, 
		   "No main function defined.  Is this a library source file?");
		return(-1);
	}
	return(eRunCode(script,&script->main));
}
