/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: util.c,v 1.19 2002/06/04 06:37:24 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>

#include "config/config.h"
#include "ember.h"
#include "parser.h"
#include "mod_str.h"
#include "mod_io.h"
#include "mod_syscall.h"
#include "mod_net.h"
#include "mod_time.h"
#include "mod_www.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/

ember	*emberScript;

int	yyparse();

/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/


/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/
void eUtilResetError(script)
	ember	*script;
{
	*script->errorMsg = 0;
}


int eInit(script)
	ember	*script;
{
	return(0);
}

int eInitMain(script)
	ember	*script;
{
	script->info.curFunction = & script->main;
	return(0);
}

int eEndOfInput()
{
	return(0);
}


void eFreeScript(script)
	ember	*script;
{
	e_funct	*curFunct,
		*prevFunct;
	int	count;


	/* Free the literal values */
	eSymFreeSymbols(script, script->literals);

	/* Free the main scope variable map */
	eSymFreeSymbolMap(script, script->main.variables);

	/* Free the main function code */
	eCodeFreeCode(script, script->main.codePtr);

	/* Free the symbol table */
	eSymFreeSymbols(script, script->variables);

	/* Free the macros */
	eSymFreeSymbolMap(script, script->macros);

	/* Free the function definitions, code, and variable maps */
	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curFunct = script->scriptFuncts[count];
		while(curFunct)
		{
			eSymFreeSymbolMap(script, curFunct->variables);
			eCodeFreeCode(script, curFunct->codePtr);
			eSymFreeArgList(script, curFunct);
			prevFunct = curFunct;
			curFunct = curFunct->next;
			free(prevFunct);
		}
		count++;
	}

	/* Free the module function structures */
	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curFunct = script->moduleFuncts[count];
		while(curFunct)
		{
			prevFunct = curFunct;
			curFunct = curFunct->next;
			free(prevFunct);
		}
		count++;
	}

	/* Free anything else */
	eFreeStackFrames();
	free(script);
}



ember *eCreateScript()
{
	ember 	*new;
	int	eDynAddModule();
	int	eSymCreateaMacro();
	int	eSymSetIntValue();
	int	eSymSetUintValue();
	int	eSymSetRealValue();
	int	eSymSetTextValue();
	void	eRunSetError();

	new = malloc(sizeof(ember));
	bzero(new,sizeof(ember));
	new->info.curFunction = &new->main;
	new->info.lexStartState = -1;

	new->methods.addDynModule = eDynAddModule;
	new->methods.createIntMacro = eSymCreateIntMacro;
	new->methods.createTextMacro = eSymCreateTextMacro;
	new->methods.createRealMacro = eSymCreateRealMacro;
	new->methods.setIntValue = eSymSetIntValue;
	new->methods.setUintValue = eSymSetUintValue;
	new->methods.setRealValue = eSymSetRealValue;
	new->methods.setTextValue = eSymSetTextValue;
	new->methods.setArrayElement = eSymSetArrayElement;
	new->methods.getArrayElement = eSymGetArrayElement;
	new->methods.getNumArrayElements = eSymGetNumArrayElements;
	new->methods.initArray = eSymInitArray;
	new->methods.freeValue = eSymFreeValue;
	new->methods.setError = eRunSetError;
	new->methods.storeObject = eStoreObject;
	new->methods.deleteObject = eDeleteObject;
	new->methods.fetchObject = eFetchObject;
	return(new);
}


int eFileSource(script,path)
	ember	*script;
	char	*path;
{
	char	*cp;

	if (path == NULL)
	{
		script->source.names[0] = strdup("standard input");
		script->source.fd = fileno(stdin);
	}
	else
	{
		cp = rindex(path,'/');
		if (cp)
			cp++;
		else
			cp = path;
		script->source.names[0] = strdup(cp);
		script->source.fd = open(path,O_RDONLY);
	}
	script->source.type = EMBER_FILE_INPUT;
	if(script->source.fd < 0)
		return(-1);
	return(0);
}



int eBufferSource(script,buf)
	ember	*script;
	char	*buf;
{
	script->source.names[0] = strdup("input buffer");
	script->source.buffer = buf;
	script->source.type = EMBER_BUFFER_INPUT;
	return(0);
}




int eParseScript(script)
	ember	*script;
{
	int		res;
	extern ember	*emberScript;
	e_stab		*sym;

        sym = eSymCreateVariable(script, "$ERRMSG");
        sym = eSymCreateVariable(script, "$EMBER_VERSION");
        eSymSetTextValue(script, &sym->value, EMBER_PROG_VERSION);
	eSymCreateTextMacro(script, "EMBER_BRK", "");
	eSymCreateTextMacro(script, "LITE_BRK", "");

	if (eAddModule(script, io_externs) < 0)
		return(-1);
	if (eAddModule(script, str_externs) < 0)
		return(-1);
	if (eAddModule(script, syscall_externs) < 0)
		return(-1);
	if (eAddModule(script, net_externs) < 0)
		return(-1);
	if (eAddModule(script, time_externs) < 0)
		return(-1);
	if (eAddModule(script, www_externs) < 0)
		return(-1);

	res = 0;
	script->info.curLineNum = 1;
	emberScript = script;
	res = yyparse();
	if (res != 0)
		return(-1);
	if (script->info.inComment)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Unterminated comment started at line %d",
			script->info.commentStart);
		return(-1);
	}
	return(0);
}


int eAddModule(script, module)
	ember	*script;
	e_extern module[];
{
	e_extern *curFunct;
	int	count;

	count = 0;
	curFunct = &module[count];
	while(curFunct->name)
	{
		if (eSymAddExternalFunction(script, curFunct) < 0)
			return(-1);
		count++;
		curFunct = &module[count];
	}
	return(0);
}

int eGenHash(name)
        char            *name;
{
        int             val;
        REG char        *cp;

        val = 0;
        cp = name;
        while( *cp)
        {
                val += *cp++;
        }
        val = val % EMBER_HASH_LEN;
        return(val);
}

void eDumpParseResult(script)
	ember	*script;
{
	e_funct	*curFunct;
	int	count;

	/*
	** Dump the global stuff
	*/
	printf("\n\nDumping internal representation of script '%s'\n\n",
		script->source.names[0]);
	printf("Literal value declarations\n");
	eSymDumpSymbols(script, script->literals);

	printf("\n\nMain scope variable declarations\n");
	eSymDumpSymbolMap(script, script->main.variables);

	printf("\n\nMain function code\n");
	eCodeDumpCode(script, script->main.codePtr);

	/*
	** Dump the functions
	*/
	count = 0;
	while(count < EMBER_HASH_LEN)
	{
		curFunct = script->scriptFuncts[count];
		while(curFunct)
		{
			printf("\n\nFunction '%s' variables\n",curFunct->name);
			eSymDumpSymbolMap(script, curFunct->variables);
			printf("\n\nFunction '%s' code\n",curFunct->name);
			eCodeDumpCode(script, curFunct->codePtr);
			curFunct = curFunct->next;
		}
		count++;
	}
}

char *eGetSourceName(script)
	ember	*script;
{
	return(script->source.names[0]);
}

char *eGetCurFileName(script)
	ember	*script;
{
	static	char resp[] = "BAD FILE ID";
	if (script->info.curFileID > script->info.lastFileID)
		return(resp);
	return(script->source.names[script->info.curFileID]);
}


int eGetLineNum(script)
	ember	*script;
{
	return(script->info.curLineNum);
}

char *eGetErrorMsg(script)
	ember	*script;
{
	return(script->errorMsg);
}


int eStoreObject(script, module, descr, objPtr)
	ember	*script;
	char	*module,
		*descr;
	void	*objPtr;
{
	int	hash,
		objId;
	e_obj	*curObj;

	script->info.lastObjectID++;
	objId = script->info.lastObjectID;
	hash = objId % EMBER_HASH_LEN;
	curObj = script->objectTable[hash];
	if (curObj == NULL)
	{
		script->objectTable[hash] = 
			(e_obj*)eMemMalloc(script, sizeof(e_obj));
		curObj = script->objectTable[hash];
	}
	else
	{
		while(curObj->next)
			curObj = curObj->next;
		curObj->next = (e_obj*)eMemMalloc(script, sizeof(e_obj));
		curObj = curObj->next;
	}
	if (curObj == NULL)
		return(-1);
	curObj->next = NULL;
	curObj->id = objId;
	curObj->obj = objPtr;
	strncpy(curObj->module, module, EMBER_NAME_LEN);
	strncpy(curObj->descr, descr, EMBER_NAME_LEN);
	return(objId);
}


void *eFetchObject(script, module, descr, objId)
	ember	*script;
	char	*module,
		*descr;
	int	objId;
{
	int	hash;
	e_obj	*curObj;

	hash = objId % EMBER_HASH_LEN;
	curObj = script->objectTable[hash];
	while(curObj)
	{
		if (curObj->id == objId)
		{
			if (strcmp(curObj->descr, descr) == 0 &&
			    strcmp(curObj->module, module) == 0 )
			{
				return(curObj->obj);
			}
			else
			{	
				return(NULL);
			}
		}
		curObj = curObj->next;
	}
	return(NULL);
}


void *eDeleteObject(script, module, descr, objId)
	ember	*script;
	char	*module,
		*descr;
	int	objId;
{
	int	hash;
	e_obj	*curObj,
		*prevObj;
	void	*ptr;

	hash = objId % EMBER_HASH_LEN;
	curObj = script->objectTable[hash];
	prevObj = NULL;
	while(curObj)
	{
		if (curObj->id == objId)
		{
			if (strcmp(curObj->descr, descr) == 0 &&
			    strcmp(curObj->module, module) == 0 )
			{
				break;
			}
			else
			{	
				return(NULL);
			}
		}
		prevObj = curObj;
		curObj = curObj->next;
	}
	if (curObj == NULL)
		return(NULL);

	/*
	** Remove the object from the list
	*/
	ptr = curObj->obj;
	if (prevObj)
		prevObj->next = curObj->next;
	else
		script->objectTable[hash] = curObj->next;
	eMemFree(script, curObj);
	return(ptr);
}


void eDebugTrap()
{
}
